use std::{fmt::Debug, hash::Hash};

use crate::{db::DBConnection, DBResult};

mod datum;
pub(crate) mod helpers;

/// Integral identifier for an entity.
pub trait EntityID: 'static + PartialEq + Hash + PartialOrd + Debug + Copy {
    type Entity: Entity<ID = Self>;

    /// Construct from a raw integer.
    ///
    /// Do not use this unless you are _very_ certain it is what you want.
    fn from_raw(id: usize) -> Self;
    fn into_raw(self) -> usize;
}

// ----------------------------------------------------------------------
// EntityDatum and related types
// ----------------------------------------------------------------------

/// Represents a data field in an Entity.
pub trait EntityDatum: 'static {
    fn sql_type() -> &'static str;

    fn bind_to<'a>(&self, _stmt: &mut sqlite::Statement<'a>, index: usize);
    fn build_from<'a>(
        conn: &DBConnection,
        ctx: &'static str,
        stmt: &mut sqlite::Statement<'a>,
        index: usize,
    ) -> DBResult<(Self, usize)>
    where
        Self: Sized;

    fn accept_entity_visitor(_: &mut impl EntityVisitor) {}
}

/// A fixed-length list of EntityDatums, usually a tuple.
pub trait EntityDatumList {
    fn accept(&self, visitor: &mut impl EntityDatumVisitor);
    fn build_from<'a>(
        conn: &DBConnection,
        ctx: &'static str,
        stmt: &mut sqlite::Statement<'a>,
    ) -> DBResult<Self>
    where
        Self: Sized;
}

/// A walker for a EntityDatumList instance.
pub trait EntityDatumVisitor {
    fn visit<ED: EntityDatum>(&mut self, datum: &ED);
}

// trait implementations for EntityDatumList
mod datum_list;

// ----------------------------------------------------------------------
// EntityPart and related types
// ----------------------------------------------------------------------

/// A single data field in an Entity, automatically derived via `#[derive(Entity)]`.
pub trait EntityPart: 'static {
    type Datum: EntityDatum;
    type Entity: Entity;

    fn part_name() -> &'static str;
    fn unique() -> bool;
}

/// Visitor for traversing all `EntityPart`s in an `Entity` or `EntityPartList`.
pub trait EntityPartVisitor {
    fn visit<EP: EntityPart>(&mut self) {}
    fn visit_datum<EP: EntityPart>(&mut self, _datum: &EP::Datum) {}
}

/// List of EntityParts.
pub trait EntityPartList: 'static {
    type DatumList: EntityDatumList;
    fn accept_part_visitor(_: &mut impl EntityPartVisitor);
    fn accept_part_visitor_ref(datum_list: &Self::DatumList, _: &mut impl EntityPartVisitor);
}

// trait implementations for EntityPartList
mod part_list;

// ----------------------------------------------------------------------
// Entity and related types
// ----------------------------------------------------------------------

/// A single database entity, aka an object type that gets its own table.
pub trait Entity: 'static {
    type Parts: EntityPartList;
    type Uniques: EntityPartList;
    type ID: EntityID<Entity = Self>;

    fn build(values: <Self::Parts as EntityPartList>::DatumList) -> Self;

    fn entity_name() -> &'static str;
    fn accept_part_visitor(visitor: &mut impl EntityPartVisitor);
    fn accept_part_visitor_ref(&self, visitor: &mut impl EntityPartVisitor);
}

/// Visitor for traversing all `Entity`s in a container.
pub trait EntityVisitor {
    fn visit<E: Entity>(&mut self);
}
