use crate::entity::helpers::check_assoc;
use crate::DBResult;
use crate::{
    entity::{
        Entity, EntityDatum, EntityDatumList, EntityDatumVisitor, EntityPart, EntityPartList,
        EntityPartVisitor,
    },
    schema::IDMap,
};
use std::hash::{Hash, Hasher};

#[derive(Hash)]
enum QueryType {
    Select(std::any::TypeId),
    Insert,
}

fn query_hash<E: Entity>(ctx: &'static str, qtype: QueryType) -> u64 {
    let mut hasher = std::collections::hash_map::DefaultHasher::new();

    qtype.hash(&mut hasher);
    std::any::TypeId::of::<E>().hash(&mut hasher);
    ctx.hash(&mut hasher);

    hasher.finish()
}

pub fn select_by<E: Entity, PL: EntityPartList>(
    map: &IDMap<E>,
    by: &PL::DatumList,
) -> DBResult<Vec<E>> {
    struct HashDatumListTypes(std::collections::hash_map::DefaultHasher);
    impl EntityDatumVisitor for HashDatumListTypes {
        fn visit<ED: EntityDatum>(&mut self, _: &ED) {
            std::any::TypeId::of::<ED>().hash(&mut self.0);
        }
    }
    let mut ty = HashDatumListTypes(Default::default());
    by.accept(&mut ty);

    map.conn().with_prepared(
        query_hash::<E>(map.ctx(), QueryType::Select(std::any::TypeId::of::<PL>())),
        || {
            let mut conditions = String::new();
            struct BuildConditions<'a>(&'a mut String);
            impl<'a> EntityPartVisitor for BuildConditions<'a> {
                fn visit<EP: EntityPart>(&mut self) {
                    if self.0.len() > 0 {
                        self.0.push_str(" and ");
                    }
                    self.0.push_str("`");
                    self.0.push_str(EP::part_name());
                    self.0.push_str("`");
                    self.0.push_str(" = ?");
                }
            }
            PL::accept_part_visitor(&mut BuildConditions(&mut conditions));

            let table_name = format!("{}_{}", map.ctx(), E::entity_name());
            format!("select rowid, * from `{}` where {}", table_name, conditions)
        },
        |stmt| {
            stmt.reset()?;
            struct BindDatum<'a, 'b>(&'a mut sqlite::Statement<'b>, usize);
            impl<'a, 'b> EntityDatumVisitor for BindDatum<'a, 'b> {
                fn visit<ED: EntityDatum>(&mut self, datum: &ED) {
                    datum.bind_to(self.0, self.1);
                    self.1 += 1
                }
            }
            // note that this indexing starts at 1
            by.accept(&mut BindDatum(stmt, 1));

            // now we grab the statement outputs
            let mut rows = vec![];
            while stmt.next()? == sqlite::State::Row {
                let datum_list =
                    <<E::Parts as EntityPartList>::DatumList as EntityDatumList>::build_from(
                        &map.conn(),
                        map.ctx(),
                        stmt,
                    )?;
                rows.push(E::build(datum_list));
            }

            Ok(rows)
        },
    )
}

pub fn insert<E: Entity>(map: &IDMap<E>, value: &E) -> DBResult<()> {
    map.conn().with_prepared(
        query_hash::<E>(map.ctx(), QueryType::Insert),
        || {
            let table_name = format!("{}_{}", map.ctx(), E::entity_name());

            let mut part_names = String::new();
            let mut placeholders = String::new();
            struct PartNameVisitor<'a>(&'a mut String, &'a mut String);
            impl<'a> EntityPartVisitor for PartNameVisitor<'a> {
                fn visit<EP: EntityPart>(&mut self) {
                    // if this is a set-association, then we don't actually want to do anything
                    // with it here; it doesn't have a column
                    if check_assoc::<EP>() {
                        return;
                    }

                    if self.0.len() != 0 {
                        self.0.push_str(", ");
                        self.1.push_str(", ");
                    }
                    self.0.push_str("`");
                    self.0.push_str(EP::part_name());
                    self.0.push_str("`");
                    self.1.push_str("?");
                }
            }

            E::accept_part_visitor(&mut PartNameVisitor(&mut part_names, &mut placeholders));

            format!(
                "insert into `{}` ({}) values ({})",
                table_name, part_names, placeholders
            )
        },
        |prepared| {
            prepared.reset()?;

            struct PartBinder<'a, 'b>(&'a mut sqlite::Statement<'b>, usize);
            impl<'a, 'b> EntityPartVisitor for PartBinder<'a, 'b> {
                fn visit_datum<EP: EntityPart>(&mut self, datum: &EP::Datum) {
                    // skip associations, as with the query preparation above
                    if check_assoc::<EP>() {
                        return;
                    }

                    datum.bind_to(self.0, self.1);
                    self.1 += 1;
                }
            }

            value.accept_part_visitor_ref(&mut PartBinder(prepared, 1));

            prepared.next()?;

            DBResult::Ok(())
        },
    )
}
