use crate::{
    db::{Connection, DBConnection},
    entity::{Entity, EntityDatum, EntityPartList, EntityVisitor},
    query,
};
use crate::{DBError, DBResult};

mod build;
mod entity;
pub(crate) mod meta;
mod tests;

// ----------------------------------------------------------------------
// Specification types
// ----------------------------------------------------------------------

pub struct IDMap<T: Entity> {
    conn: DBConnection,
    ctx: &'static str,
    _ghost: std::marker::PhantomData<T>,
}

impl<T: Entity> IDMap<T> {
    pub fn build(db: DBConnection, ctx: &'static str) -> Self {
        Self {
            conn: db,
            ctx,
            _ghost: std::marker::PhantomData,
        }
    }

    pub(crate) fn conn(&self) -> &DBConnection {
        &self.conn
    }

    pub(crate) fn ctx(&self) -> &'static str {
        self.ctx
    }

    pub fn by_id(&self, _id: T::ID) -> DBResult<Option<T>> {
        // query::select_by
        todo!()
    }

    pub fn lookup_unique(
        &self,
        uniques: &<<T as Entity>::Uniques as EntityPartList>::DatumList,
    ) -> DBResult<Option<T>> {
        query::select_by::<T, T::Uniques>(self, uniques).map(|mut v| {
            if v.len() > 0 {
                Some(v.remove(0))
            } else {
                None
            }
        })
    }

    pub fn insert(&self, value: &T) -> DBResult<()> {
        query::insert(self, value)
    }
}

pub struct Index<T: Entity, Key: EntityDatum> {
    _conn: DBConnection,
    _ghost: std::marker::PhantomData<(T, Key)>,
}

pub(crate) struct AssocData {
    pub(crate) conn: DBConnection,
    pub(crate) ctx: &'static str,
    pub(crate) base_rowid: Option<i64>,
}

pub struct AssocSet<T: Entity> {
    data: Option<AssocData>,
    _ghost: std::marker::PhantomData<T>,
}

impl<T: Entity> AssocSet<T> {
    pub fn empty() -> Self {
        Self {
            data: None,
            _ghost: Default::default(),
        }
    }
}

impl<T: Entity> EntityDatum for AssocSet<T> {
    fn sql_type() -> &'static str {
        unreachable!()
    }

    fn accept_entity_visitor(v: &mut impl EntityVisitor) {
        v.visit::<T>();
    }

    fn bind_to<'a>(&self, _stmt: &mut sqlite::Statement<'a>, _index: usize) {
        unreachable!()
    }

    fn build_from<'a>(
        conn: &DBConnection,
        ctx: &'static str,
        stmt: &mut sqlite::Statement<'a>,
        index: usize,
    ) -> DBResult<(Self, usize)>
    where
        Self: Sized,
    {
        // assuming that the stmt has the rowid as index 0
        Ok((
            Self {
                data: Some(AssocData {
                    conn: conn.clone(),
                    ctx,
                    base_rowid: stmt.read(0)?,
                }),
                _ghost: Default::default(),
            },
            index,
        ))
    }
}

pub trait DatabaseItem {
    fn item_key() -> &'static str;
    fn dependency_keys() -> &'static [&'static str];

    fn is_index() -> bool {
        false
    }
    fn index_over() -> &'static str {
        unreachable!()
    }
    fn index_columns() -> &'static [&'static str] {
        unreachable!()
    }

    fn accept_entity_visitor(visitor: &mut impl EntityVisitor);
}

pub trait DatabaseItemVisitor {
    fn visit<DI: DatabaseItem>(&mut self)
    where
        Self: Sized;
}

pub trait DatabaseSpec {
    fn accept_entity_visitor(visitor: &mut impl EntityVisitor);
}

impl<T: Entity> DatabaseSpec for IDMap<T> {
    fn accept_entity_visitor(visitor: &mut impl EntityVisitor) {
        visitor.visit::<T>()
    }
}

pub trait Database {
    fn open_path<U: AsRef<str>>(uri: U) -> Result<Self, DBError>
    where
        Self: Sized,
    {
        let conn = Connection::open(uri)?;
        let schema = build::collect_from_database::<Self>();
        if !schema.check(conn.clone()) {
            schema.create(conn.clone())
        }

        let db = Self::build(conn);

        Ok(db)
    }

    #[doc(hidden)]
    fn build(conn: DBConnection) -> Self
    where
        Self: Sized;

    fn accept_item_visitor(visitor: &mut impl DatabaseItemVisitor)
    where
        Self: Sized;
}
